/*
 * This file belongs to FreeMiNT.  It's not in the original MiNT 1.12
 * distribution.  See the file Changes.MH for a detailed log of changes.
 */

/*
 * This file is dedicated to the FreeMiNT project.
 * It's not allowed to use this file for other projects without my
 * explicit permission.
 */

/*
 * begin:	1999-01-05
 * last change: 1999-04-13
 * 
 * Author: Frank Naumann - <fnaumann@cs.uni-magdeburg.de>
 * 
 * please send suggestions or bug reports to me or
 * the MiNT mailing list
 * 
 * 
 * changes since last version:
 * 
 * 
 * 
 * known bugs:
 * 
 * - nothing
 * 
 * todo:
 * 
 * 
 * explanation:
 * 
 * - every DMA driver have first register into the kernel;
 *   the driver get then a unique channel number for his DMA
 * - the kernel allocate also a semaphore for this DMA channel
 *   and will automatically block all processes that will
 *   access this channel [semaphore]
 * - before using a DMA channel the driver must first reserve
 *   the channel [semaphore] with dma_start(channel)
 * - if the semaphore is locked (a DMA transfer is already in progress)
 *   the kernel block all processes until the semaphore is unlocked
 * - then the driver can setup the DMA transfer, program the interrupt
 *   and so on
 * - if this stuff is done the driver call dma_block(channel) that will block
 *   the actual process and give up the processor
 * - now another process can run; if these process also access these DMA channel
 *   it will automatically blocked because the semaphore is locked
 * - ok, sometimes the Interrupt occurs (hopefully :-))
 * - the interrupt get the driver and not the kernel
 * - now the driver can finish his operation, cleanup and so on in his
 *   Interrupt service routine
 * - before the Interrupt service routine jump back it must call
 *   dma_deblock(channel) that will wake up all processes which wait for the
 *   semaphore; dma_deblock(channel) also unlock the semaphore
 */

# include "dma.h"
# include "proc.h"


# define DMA_VERS	1		/* internal version */

/****************************************************************************/
/* BEGIN definition part */

/*
 * internal prototypes
 */

/* END definition part */
/****************************************************************************/

/****************************************************************************/
/* BEGIN global data */

# define CHANNELS	32
static struct
{
	ushort	used;
	ushort	lock;
	ushort	pid;
	
} channels [CHANNELS];

/* END global data */
/****************************************************************************/

/****************************************************************************/
/* BEGIN  */

ushort _cdecl
dma_register (void)
{
	long i;
	
	for (i = 0; i < CHANNELS; i++)
	{
		if (!(channels [i].used))
		{
			channels [i].used = 1;
			channels [i].lock = 0;
			channels [i].pid = -1;
			
			return i;
		}
	}
	
	return 0xffff;
}

void _cdecl
dma_unregister (ushort c)
{
	if (channel < CHANNELS && channels [c].used)
		channels [c].used = 0;
}

# if 1
/* only synchronus callable [to kernel]
 */
void _cdecl
dma_start (ushort c)
{
	while (channels [c].lock)
	{
		channels [c].pid = curproc->pid;
		sleep (0x100 | IO_Q, &(channels [c]));
		channels [c].pid = -1;
	}
	
	channels [c].lock = 1;
}

void _cdecl
dma_block (ushort c)
{
	channels [c].pid = curproc->pid;
	sleep (0x100 | IO_Q, &(channels [c]));
	channels [c].pid = -1;
}
# else

void _cdecl
dma_block (ushort c)
{
	while (channels [c].lock)
	{
		channels [c].pid = curproc->pid;
		sleep (0x100 | IO_Q, &(channels [c]));
		channels [c].pid = -1;
	}
	
	channels [c].lock = 1;
	
	channels [c].pid = curproc->pid;
	
	-> packe Ruecksprungadresse von naechster Anweisung auf den Stack
	-> rts
	-> zurueck im treiber
	-> setzt DMA auf
	-> rts
	-> sleep (0x100 | IO_Q, &(channels [c]));
	-> channels [c].pid = -1;
}

# endif

/* synchronus/asynchronus callable
 */
void _cdecl
dma_deblock (ushort channel)
{
	iwake (IO_Q, &(channels [c]), channels [c].pid);
	channels [c].lock = 0;
}

/* END  */
/****************************************************************************/

/****************************************************************************/
/* BEGIN  */

/* END  */
/****************************************************************************/
