/*
 * This file belongs to FreeMiNT.  It's not in the original MiNT 1.12
 * distribution.  See the file Changes.MH for a detailed log of changes.
 */

/*
 * This file is dedicated to the FreeMiNT project.
 * It's not allowed to use this file for other projects without my
 * explicit permission.
 */

/*
 * begin:	1998-06
 * last change: 1998-09-10
 * 
 * Author: Frank Naumann - <fnaumann@cs.uni-magdeburg.de>
 * 
 * please send suggestions, patches or bug reports to me or
 * the MiNT mailing list
 * 
 */

# ifndef _BLOCK_IO_H
# define _BLOCK_IO_H

# ifdef __TURBOC__
# include "include\misc.h"
# include "include\types.h"
# else
# include "misc.h"
# include "types.h"
# endif


/*
 * definitions
 */

typedef struct di DI;
typedef struct unit UNIT;
typedef struct cbl CBL;

/* device identifikator */
struct di
{
	DI	*next;			/* internal: next in linked list */
	UNIT	**table;		/* internal: unit hash table */
	UNIT	*wb_queue;		/* internal: writeback queue */
	
	ushort	drv;			/* internal: BIOS device number (unique) */
	ushort	major;			/* XHDI */
	ushort	minor;			/* XHDI */
	ushort	mode;			/* internal: some flags */
	
# define BIO_WP_MODE		0x01	/* write-protect bit (soft) */
# define BIO_WB_MODE		0x02	/* write-back bit (soft) */
# define BIO_REMOVABLE		0x04	/* removable media */
# define BIO_LOCKABLE		0x08	/* lockable media */
# define BIO_LRECNO		0x10	/* lrecno supported */
# define BIO_WP_HARD		0x20	/* write-protected partition */
	
	ulong	start;			/* physical start sector */
	ulong	size;			/* physical sectors */
	ulong	pssize;			/* internal: physical sector size */
	
	ushort	pshift;			/* internal: size to count calculation */
	ushort	lshift;			/* internal: logical to physical recno calculation */
	
	long	_cdecl (*rwabs)(DI *di, ushort rw, void *buf, ulong size, ulong lrecno);
	long	_cdecl (*dskchng)(DI *di);
	
	ushort	valid;			/* internal: DI valid */
	ushort	lock;			/* internal: DI in use */
	
	char	id[4];			/* XHDI partition id (GEM, BGM, RAW, \0D6, ...) */
	ushort	key;			/* XHDI key */
	
	char	res[18];		/* reserved for future */
};

/* cache unit */
struct unit
{
	char	*data;			/* pointer to the data */
	UNIT	*next;			/* next in hash table */
	UNIT	*wb_prev;		/* prev in writeback queue */
	UNIT	*wb_next;		/* next in writeback queue */
	CBL	*cbl;			/* the parent cache block or NULL for resident UNITs */
	DI	*di;			/* device identifikator */
	ulong	sector;			/* start sector on medium */
	ulong	size;			/* size in bytes */
	ulong	stat;			/* access statistic */
	ushort	pos;			/* location in block */
	ushort	dirty;			/* unit is modified (flag) */
	ushort	lock;			/* unit is locked (counter) */
	ushort	res;			/* reserved for future */	
};


/*
 * interface
 */

typedef struct
{
	ushort	version;		/* buffer cache version */
	ushort	revision;		/* buffer cache revision */
	
# define BLOCK_IO_VERS	3		/* our existing version - incompatible interface change */
# define BLOCK_IO_REV	1		/* actual revision - compatible interface change */
	
	long	_cdecl (*config)	(const ushort drv, const long config, const long mode);
	
/* config: */
# define BIO_WP		1		/* configuring writeprotect feature */
# define BIO_WB		2		/* configuring writeback mode */
# define BIO_MAX_BLOCK	10		/* return maximum cacheable blocksize */
# define BIO_DEBUGLOG	100		/* only for debugging, kernel internal */
# define BIO_DEBUG_T	101		/* only for debugging, kernel internal */
	
	/* DI management */
	DI *	_cdecl (*get_di)	(ushort drv);
	DI *	_cdecl (*res_di)	(ushort drv);
	void	_cdecl (*free_di)	(DI *di);
	
	/* physical/logical mapping setting */
	void	_cdecl (*set_pshift)	(DI *di, ulong physical);
	void	_cdecl (*set_lshift)	(DI *di, ulong logical);
	
	/* cached block I/O */
	UNIT *	_cdecl (*lookup)	(DI *di, ulong sector, ulong blocksize);
	UNIT *	_cdecl (*getunit)	(DI *di, ulong sector, ulong blocksize);
	UNIT *	_cdecl (*read)		(DI *di, ulong sector, ulong blocksize);
	long	_cdecl (*write)		(UNIT *u);
	long	_cdecl (*l_read)	(DI *di, ulong sector, ulong blocks, ulong blocksize, void *buf);
	long	_cdecl (*l_write)	(DI *di, ulong sector, ulong blocks, ulong blocksize, void *buf);
	
	/* optional feature */
	void	_cdecl (*pre_read)	(DI *di, ulong *sector, ulong blocks, ulong blocksize);
	
	/* synchronization */
	void	_cdecl (*lock)		(UNIT *u);
	void	_cdecl (*unlock)	(UNIT *u);
	
	/* update functions */
	void	_cdecl (*mark_modified)	(UNIT *u);
	void	_cdecl (*sync_drv)	(DI *di);
	
	/* cache management */
	long	_cdecl (*validate)	(DI *di, ulong maxblocksize);
	void	_cdecl (*invalidate)	(DI *di);
	
	/* revision 1 extension: resident block I/O
	 */
	UNIT *	_cdecl (*get_resident)	(DI *di, ulong sector, ulong blocksize);
	void	_cdecl (*rel_resident)	(UNIT *u);
	
	/* revision 2 extension: remove explicitly a cache unit without writing
	 */
	void	_cdecl (*remove)	(UNIT *u);
	
	
	long	res[3];			/* reserved for future */
} BIO;


/*
 * useful makros
 */

# define BIO_RWABS(di, rw, buf, size, recno) \
				((*(di->rwabs))(di, rw, buf, size, recno))
# define BIO_DSKCHNG(di)	((*(di->dskchng))(di))

# define BIO_WP_CHECK(di)	((di->mode & BIO_WP_MODE) ? ENABLE : DISABLE)
# define BIO_WB_CHECK(di)	((di->mode & BIO_WB_MODE) ? ENABLE : DISABLE)
# define BIO_LR_CHECK(di)	((di->mode & BIO_LRECNO) ? ENABLE : DISABLE)

FASTFN void
bio_MARK_MODIFIED (BIO *bio, UNIT *u)
{
	if (!u->dirty)
	{
		(bio)->mark_modified (u);
	}
}

FASTFN void
bio_SYNC_DRV (BIO *bio, DI *di)
{
	if (!BIO_WB_CHECK (di))
	{
		(bio)->sync_drv (di);
	}
}


# endif /* _BLOCK_IO_H */
