/*
 * This file has been modified as part of the FreeMiNT project. See
 * the file Changes.MH for details and dates.
 */

/*
 * Copyright 1990,1991,1992 Eric R. Smith.
 * Copyright 1992,1993,1994 Atari Corporation.
 * All rights reserved.
 */

# include "main.h"
# include "global.h"

# include "bios.h"	/* init_bios, bflush */
# include "block_IO.h"	/* init_block_IO */
# include "cnf.h"	/* load_config, some variables */
# include "console.h"
# include "cookie.h"
# include "detect.h"
# include "dos.h"
# include "dosdir.h"
# include "dosfile.h"
# include "dosmem.h"
# include "filesys.h"	/* init_filesys, s_ync, load_*, close_filesys */
# include "kernel.h"	/* enter_gemdos */
# include "memory.h"	/* init_mem, get_region, attach_region, restr_screen */
# include "proc.h"	/* init_proc, add_q, rm_q */
# include "signal.h"	/* post_sig */
# include "string.h"
# include "timeout.h"
# include "util.h"
# include "welcome.h"
# include "xbios.h"

# include "version.h"


/* the kernel's stack size */
# define STACK	8 * 1024L

/* if the user is holding down the magic shift key, we ask before booting */
# define MAGIC_SHIFT	0x2		/* left shift */

/* magic number to show that we have captured the reset vector */
# define RES_MAGIC	0x31415926L

static long getmch (void);
# ifndef SYSUPDATE_DAEMON
static void do_sync (void);
# endif
void shutdown (void);
static long _cdecl mint_criticerr (long);
static void _cdecl do_exec_os (register long basepage);

static int gem_active;		/* 0 if AES has not started, nonzero otherwise */

COOKIE	*clock_cookie_p;	/* used in time.c */

# define EXEC_OS 0x4feL
static int	check_for_gem (void);
static void	run_auto_prgs (void);
static int	boot_kernel_p (void);


/* print a additional boot message
 */

void
boot_print (const char *s)
{
	Cconws (s);	
}

# include <stdarg.h>

void
boot_printf (const char *fmt, ...)
{
	char buf [SPRINTF_MAX];
	va_list args;
	
	va_start (args, fmt);
	vsprintf (buf, fmt, args);
	va_end (args);
	
	Cconws (buf);
}


# ifdef LATTICE
/*
 * AGK: this is witchcraft to completely replace the startup code for
 * Lattice; doing so saves around 10K on the final binary and pulls only
 * long division & multiplication from the library (and not even those
 * if you compile for native '030). The drawback of this code is it
 * passes no environment or command line whatsoever. Since I always
 * set MiNT options & environment in 'mint.cnf' this is not a personal
 * downer, however at some point in the future we ought to have a kernel
 * parseargs() like call which sets these things up.
 */ 
BASEPAGE *_base;

static void
start(BASEPAGE *bp)
{
	long shrinklen;
	
	_base = bp;
	shrinklen = bp->p_tlen + bp->p_dlen + bp->p_blen + STACK + 0x100;
	if (bp->p_lowtpa + shrinklen <= bp->p_hitpa)
	{
		static char null[1] = {""};
		static char *argv[2] = {null, NULL};
		extern __builtin_putreg P_((int, long));	/* totally bogus */

		__builtin_putreg(15, bp->p_lowtpa + shrinklen);
		Mshrink((void *)bp->p_lowtpa, shrinklen);
		main(1, argv);
	}
	Pterm(ENSMEM);
}
# endif

# if defined(__GNUC__) || defined(__MINT__)
long _stksize = STACK;
# ifndef PROFILING
# include <minimal.h>
# endif
# endif

int curs_off = 0;	/* set if we should turn the cursor off when exiting */
int mint_errno = 0;	/* error return from open and creat filesystem calls */

/*
 * AGK: for proper co-processors we must consider saving their context.
 * This variable when non-zero indicates that the BIOS considers a true
 * coprocessor to be present. We use this variable in the context switch
 * code to decide whether to attempt an FPU context save.
 */
short fpu = 0;		/* flag */
long fputype = 0;	/* value for cookie jar */

/*
 * "mch" holds what kind of machine we are running on
 */
long mch = 0;

/*
 * "screen_boundary+1" tells us how screens must be positioned
 * (to a 256 byte boundary on STs, a 16 byte boundary on other
 * machines; actually, 16 bytes is conservative, 4 is probably
 * OK, but it doesn't hurt to be cautious). The +1 is because
 * we're using this as a mask in the ROUND() macro in mem.h.
 */
int screen_boundary = 255;

/*
 * variable holds processor type
 */
long mcpu = 0;

/*
 * variable holds language preference
 */
int gl_lang = -1;

/*
 * variable set if someone has already installed an flk cookie
 */
int flk = 0;

/*
 * variable set to 1 if the _VDO cookie indicates Falcon style video
 */
int FalconVideo = 0;

/*
 * variable set to 1 if the _VDO cookie indicates STE style video
 */
short ste_video = 0;

/* 
 * when memory protection is not used vm is switched on.
 * vm is switched off if not 030
 * (current implementation can not handle 68851,040 or 060 mmus)
 */

# ifdef VM_EXTENSION
int vm_in_use=0; /* true if vm is used */
ulong st_left_after_vm = 0x00400000L;
# endif

/*
 * if this variable is set, then we have "secure mode" enabled; that
 * means only the superuser may access some functions those may be critical
 * for system security.
 *
 */

int secure_mode = 0;

/* in mem.c */

extern short forcefastload;

# ifndef MULTITOS
BASEPAGE *gem_base = 0;
long gem_start = 0;
# endif

/* GEMDOS pointer to current basepage */
BASEPAGE **tosbp;

/* pointer to the BIOS keyboard shift variable */
extern char *kbshft;		/* see bios.c */
extern BCONMAP2_T *bconmap2;	/* bconmap struct, see bios.c */

/* version of TOS we're running over */
int tosvers;

/* version of GEMDOS in ROM.
 * dos.c no longer directly calls Sversion()
 */

#ifdef OLDTOSFS
long gemdos_version;
#endif

/* flag for Falcon TOS kludges: TRUE if TOS 4.00 - 4.04
 * what about TOS 4.50 in Milan? see syscall.spp
 */
short falcontos;

/* structures for keyboard/MIDI interrupt vectors */
KBDVEC *syskey, oldkey;
extern long old_ikbd;			/* old ikbd vector */

/* values the user sees for the DOS, BIOS, and XBIOS vectors */
# if 0
long save_dos, save_bios, save_xbios;
# endif

/* values for original system vectors */
extern long old_dos, old_bios, old_xbios;
xbra_vec old_criticerr;
xbra_vec old_execos;

/* interrupt vectors are linked some other way, see new_xbra_install() */
extern long old_timer, old_vbl, old_5ms;

long old_term;

extern long old_resvec;	/* old reset vector */
long old_resval;	/* old reset validation */

# ifdef EXCEPTION_SIGS
/* bus error, address error, illegal instruction, etc. vectors */
xbra_vec old_bus, old_addr, old_ill, old_divzero, old_trace, old_priv;
xbra_vec old_linef, old_chk, old_trapv, old_mmuconf, old_format, old_cpv;
xbra_vec old_uninit, old_spurious, old_fpcp[7], old_pmmuill, old_pmmuacc;
# endif

/* BIOS disk vectors */
extern long old_mediach, old_getbpb, old_rwabs;

/* BIOS drive map */
long olddrvs;

/* interval for syncing the filesystems */
# ifdef SYSUPDATE_DAEMON
extern long sync_time;
# else
long sync_time = 5;
# endif

/* table of processor frame sizes in _words_ (not used on MC68000) */
uchar framesizes[16] =
{
	/*0*/	0,	/* MC68010/M68020/M68030/M68040 short */
	/*1*/	0,	/* M68020/M68030/M68040 throwaway */
	/*2*/	2,	/* M68020/M68030/M68040 instruction error */
	/*3*/	2,	/* M68040 floating point post instruction */
	/*4*/	3,	/* MC68LC040/MC68EC040 unimplemented floating point instruction */
			/* or */
			/* MC68060 access error */
	/*5*/	0,	/* NOTUSED */
	/*6*/	0,	/* NOTUSED */
	/*7*/	26,	/* M68040 access error */	
	/*8*/	25,	/* MC68010 long */	
	/*9*/	6,	/* M68020/M68030 mid instruction */
	/*A*/	12,	/* M68020/M68030 short bus cycle */
	/*B*/	42,	/* M68020/M68030 long bus cycle */
	/*C*/	8,	/* CPU32 bus error */
	/*D*/	0,	/* NOTUSED */
	/*E*/	0,	/* NOTUSED */
	/*F*/	13	/* 68070 and 9xC1xx microcontroller address error */
};

/* TOS and MiNT cookie jars, respectively. See the comments and code 
 * after main() for further details
 */

COOKIE *oldcookie, *newcookie;
MEMREGION *newjar_region;

/*
 * install a new vector for address "addr", using the XBRA protocol.
 * must run in supervisor mode!
 */

/*
	WHO DID THIS?
	self-modifying code is a bad idea - especially when not flushing
	CPU caches afterwards!
	
	The best idea would be to get rid of the separate struct xbra,
	and instead have enough space in front of all routines that might be
	installed in a XBRA chain, so we can just patch in the XBRA chain
	vector. This would also get rid of one unneccessary jump (which
	flushes the instruction pipe on '040).
*/

static void
xbra_install (xbra_vec *xv, long addr, long _cdecl (*func) ())
{
	xv->xbra_magic = XBRA_MAGIC;
	xv->xbra_id = MINT_MAGIC;
	xv->jump = JMP_OPCODE;
	xv->this = func;
	xv->next = *((struct xbra **) addr);
	*((short **) addr) = &xv->jump;
	cpush (xv, sizeof (xv)); 	/* workaround for now - ms */
}

/* new XBRA installer; for now only used on VBL, 200Hz and GEMDOS timer
 * but other stuff should be subsequently made linked by this
 * instead of by the xbra_install()
 */

static void
new_xbra_install (long *xv, long addr, long _cdecl (*func) ())
{
	*xv = *(long *) addr;
	*(long *) addr = (long) func;
	cpush ((long *) addr, sizeof (addr));	/* better to be safe... */
	cpush (xv, sizeof (xv));
}

/*
 * MiNT critical error handler; all it does is to jump through
 * the vector for the current process
 */

static long _cdecl
mint_criticerr (long error) /* high word is error, low is drive */
{
	return (*curproc->criticerr)(error);
}

/*
 * if we are MultiTOS, and if we are running from the AUTO folder,
 * then we grab the exec_os vector and use that to start GEM; that
 * way programs that expect exec_os to act a certain way will still
 * work.
 * NOTE: we must use Pexec instead of p_exec here, because we will
 * be running in a user context (that of process 1, not process 0)
 */

static void _cdecl
do_exec_os (long basepage)
{
	register long r;

	/* if the user didn't specify a startup program, jump to the ROM */
	if (!init_prg)
	{
		register void _cdecl (*f) P_((long));
		f = (void _cdecl (*) P_((long))) old_execos.next;
		(*f)(basepage);
		Pterm0();
	}
	else
	{
		/* we have to set a7 to point to lower in our TPA;
		 * otherwise we would bus error right after the Mshrink call!
		 */
		setstack(basepage+500L);
# if defined(__TURBOC__) && !defined(__MINT__)
		Mshrink(0, (void *)basepage, 512L);
# else
		Mshrink((void *)basepage, 512L);
# endif
		r = Pexec(200, (char *)init_prg, init_tail, init_env);
		Pterm((int)r);
	}
}


/*
 * initialize all interrupt vectors and new trap routines
 * we also get here any TOS variables that we're going to change
 * (e.g. the pointer to the cookie jar) so that rest_intr can
 * restore them.
 */

static void
init_intr (void)
{
	extern long _cdecl unused_trap();
	extern long _cdecl mint_bios();
	extern long _cdecl mint_dos();
	extern long _cdecl mint_timer();
	extern long _cdecl mint_vbl();
	extern long _cdecl mint_5ms();
	extern long _cdecl mint_xbios();
	extern long _cdecl reset();
  	extern long _cdecl new_ikbd();
  	extern long _cdecl new_bus();
  	extern long _cdecl new_addr();
  	extern long _cdecl new_ill();
  	extern long _cdecl new_divzero();
  	extern long _cdecl new_trace();
  	extern long _cdecl new_priv();
  	extern long _cdecl new_linef();
  	extern long _cdecl new_chk();
  	extern long _cdecl new_trapv();
  	extern long _cdecl new_fpcp();
  	extern long _cdecl new_mmu();
  	extern long _cdecl new_format();
  	extern long _cdecl new_cpv();
  	extern long _cdecl new_uninit();
  	extern long _cdecl new_spurious();
  	extern long _cdecl new_pmmuacc();
	extern long intr_shadow;
	
	long dummy;
	short savesr;
	int i;
	
	syskey = (KBDVEC *)Kbdvbase();
	oldkey = *syskey;

	new_xbra_install(&old_ikbd, (long)(&syskey->ikbdsys), new_ikbd);

	/* gratuitous (void *) for Lattice */
	old_term = (long)Setexc(0x102, (void (*)())-1UL);

	savesr = spl7();

	/* Take all traps; notice, that the "old" address for any trap
	 * except #1, #13 and #14 (GEMDOS, BIOS, XBIOS) is lost.
	 * It does not matter, because MiNT does not pass these
	 * calls along anyway.
	 * The main problem is, that activating this makes ROM VDI
	 * no more working, and actually any program that takes a
	 * trap before MiNT is loaded.
	 * WARNING: NVDI 5.00 uses trap #15 and isn't even polite
	 * enough to link it with XBRA.
	 */

# ifdef TRAPS_PRIVATE
	new_xbra_install(&dummy, 0x80L, unused_trap);		/* trap #0 */
	new_xbra_install(&old_dos, 0x84L, mint_dos);		/* trap #1, GEMDOS */
# if 0
	new_xbra_install(&dummy, 0x88L, unused_trap);		/* trap #2, GEM */
# endif
	new_xbra_install(&dummy, 0x8cL, unused_trap);		/* trap #3 */
	new_xbra_install(&dummy, 0x90L, unused_trap);		/* trap #4 */
	new_xbra_install(&dummy, 0x94L, unused_trap);		/* trap #5 */
	new_xbra_install(&dummy, 0x98L, unused_trap);		/* trap #6 */
	new_xbra_install(&dummy, 0x9cL, unused_trap);		/* trap #7 */
	new_xbra_install(&dummy, 0xa0L, unused_trap);		/* trap #8 */
	new_xbra_install(&dummy, 0xa4L, unused_trap);		/* trap #9 */
	new_xbra_install(&dummy, 0xa8L, unused_trap);		/* trap #10 */
	new_xbra_install(&dummy, 0xacL, unused_trap);		/* trap #11 */
	new_xbra_install(&dummy, 0xb0L, unused_trap);		/* trap #12 */
	new_xbra_install(&old_bios, 0xb4L, mint_bios);		/* trap #13, BIOS */
	new_xbra_install(&old_xbios, 0xb8L, mint_xbios);	/* trap #14, XBIOS */
# if 0
	new_xbra_install(&dummy, 0xbcL, unused_trap);		/* trap #15 */
# endif
# else
	new_xbra_install(&old_dos, 0x84L, mint_dos);		/* trap #1, GEMDOS */
	new_xbra_install(&old_bios, 0xb4L, mint_bios);		/* trap #13, BIOS */
	new_xbra_install(&old_xbios, 0xb8L, mint_xbios);	/* trap #14, XBIOS */
# endif

	new_xbra_install(&old_timer, 0x400L, mint_timer);
	xbra_install(&old_criticerr, 0x404L, mint_criticerr);
	new_xbra_install(&old_5ms, 0x114L, mint_5ms);
	new_xbra_install(&old_vbl, 4*0x1cL, mint_vbl);
	
	new_xbra_install(&old_resvec, 0x042aL, reset);
	old_resval = *((long *)0x426L);
	*((long *)0x426L) = RES_MAGIC;
	
	spl(savesr);

	/* set up signal handlers */
	xbra_install(&old_bus, 8L, new_bus);
# ifdef EXCEPTION_SIGS
	xbra_install(&old_addr, 12L, new_addr);
	xbra_install(&old_ill, 16L, new_ill);
	xbra_install(&old_divzero, 20L, new_divzero);
	xbra_install(&old_trace, 36L, new_trace);
	xbra_install(&old_priv, 32L, new_priv);
	if (tosvers >= 0x106)
		xbra_install(&old_linef, 44L, new_linef);
	xbra_install(&old_chk, 24L, new_chk);
	xbra_install(&old_trapv, 28L, new_trapv);
	for (i = (int)(sizeof(old_fpcp) / sizeof(old_fpcp[0])); i--; )
	{
		xbra_install(&old_fpcp[i], 192L + i * 4, new_fpcp);
	}
	xbra_install(&old_mmuconf, 224L, new_mmu);
	xbra_install(&old_pmmuill, 228L, new_mmu);
	xbra_install(&old_pmmuacc, 232L, new_pmmuacc);
	xbra_install(&old_format, 56L, new_format);
	xbra_install(&old_cpv, 52L, new_cpv);
	xbra_install(&old_uninit, 60L, new_uninit);
	xbra_install(&old_spurious, 96L, new_spurious);
# endif
	
	/* set up disk vectors */
	new_xbra_install(&old_mediach, 0x47eL, new_mediach);
	new_xbra_install(&old_rwabs, 0x476L, new_rwabs);
	new_xbra_install(&old_getbpb, 0x472L, new_getbpb);
	olddrvs = *((long *)0x4c2L);

	/* set up cookie jar */
	oldcookie = *CJAR;	/* CJAR defined in cookie.h */
	install_cookies();
	
	/* initialize psigintr() call stuff (on 68020+ only). */
# ifndef ONLY030
	if (mcpu > 10)
# endif
	{
		intr_shadow = (long) attach_region (rootproc,
				get_region (core, 1024, PROT_S));
		if (intr_shadow)
			quickmove((char *) intr_shadow, (char *) 0x0L, 1024L);
	}
}

/* restore all interrupt vectors and trap routines
 *
 * NOTE: This is *not* the approved way of unlinking XBRA trap handlers.
 * Normally, one should trace through the XBRA chain. However, this is
 * a very unusual situation: when MiNT exits, any TSRs or programs running
 * under MiNT will no longer exist, and so any vectors that they have
 * caught will be pointing to never-never land! So we do something that
 * would normally be considered rude, and restore the vectors to
 * what they were before we ran.
 * BUG: we should restore *all* vectors, not just the ones that MiNT caught.
 */

extern void Reset_mmu(void);		/* in vm.s */

void
restr_intr (void)
{
	short savesr;
	int i;

	savesr = spl7();
	*syskey = oldkey;		/* restore keyboard vectors */
	*tosbp = _base;			/* restore GEMDOS basepage pointer */
	*CJAR = oldcookie;		/* restore old cookie jar */

	*((long *)0x08L) = (long) old_bus.next;
# ifdef EXCEPTION_SIGS
	*((long *)0x0cL) = (long) old_addr.next;
	*((long *)0x10L) = (long) old_ill.next;
	*((long *)0x14L) = (long) old_divzero.next;
	*((long *)0x20L) = (long) old_priv.next;
	*((long *)0x24L) = (long) old_trace.next;
	if (old_linef.next)
		*((long *)0x2cL) = (long) old_linef.next;
	*((long *)0x18L) = (long) old_chk.next;
	*((long *)0x1cL) = (long) old_trapv.next;
	for (i = (int)(sizeof(old_fpcp) / sizeof(old_fpcp[0])); i--; )
	{
		((long *)0xc0L)[i] = (long) old_fpcp[i].next;
	}
	*((long *)0xe0L) = (long) old_mmuconf.next;
	*((long *)0xe4L) = (long) old_pmmuill.next;
	*((long *)0xe8L) = (long) old_pmmuacc.next;
	*((long *)0x38L) = (long) old_format.next;
	*((long *)0x34L) = (long) old_cpv.next;
	*((long *)0x3cL) = (long) old_uninit.next;
	*((long *)0x60L) = (long) old_spurious.next;
# endif
	*((long *)0x84L) = old_dos;
	*((long *)0xb4L) = old_bios;
	*((long *)0xb8L) = old_xbios;
	*((long *)0x408L) = old_term;
	*((long *)0x404L) = (long) old_criticerr.next;
	*((long *)0x114L) = old_5ms;
	*((long *)0x400L) = old_timer;
	*((long *)0x70L) = old_vbl;
	*((long *)0x426L) = old_resval;
	*((long *)0x42aL) = old_resvec;
	*((long *)0x476L) = old_rwabs;
	*((long *)0x47eL) = old_mediach;
	*((long *)0x472L) = old_getbpb;
	*((long *)0x4c2L) = olddrvs;
# ifdef VM_EXTENSION
	if (vm_in_use)
	{
		Reset_mmu();
	}
# endif
	spl(savesr);
}


/* we save the TOS supervisor stack pointer so that we can reset it when
 * calling Pterm() (not that anyone will ever want to leave MiNT :-)).
 */

long tosssp;		/* TOS supervisor stack pointer */


# ifndef SYSUPDATE_DAEMON
/*
 * do_sync: sync all filesystems at regular intervals
 */
static void
do_sync (void)
{
	s_ync ();
	addroottimeout (1000L * sync_time, (void _cdecl (*)(PROC *p)) do_sync, 0);
}
# endif


/*
 * shut down processes; this involves waking them all up, and sending
 * them SIGTERM to give them a chance to clean up after themselves
 */

static void _cdecl
shutmedown (PROC *p)
{
	UNUSED(p);
	curproc->wait_cond = 0;
}

void
shutdown (void)
{
	PROC *p;
	int proc_left = 0;

	curproc->sighandle[SIGCHLD] = SIG_IGN;

	for (p = proclist; p; p = p->gl_next)
	{
		if (p->pid == 0) continue;
		if (p->wait_q != ZOMBIE_Q && p->wait_q != TSR_Q)
		{
			short sr = spl7();
			if (p->wait_q != READY_Q)
			{
				rm_q(p->wait_q, p);
				add_q(READY_Q, p);
			}
			spl(sr);
			post_sig(p, SIGTERM);
			proc_left++;
		}
	}

	if (proc_left)
	{
		/* sleep a little while, to give the other processes a chance to
		 * shut down
		 */
		addtimeout(5000, shutmedown);
		do {
			sleep(WAIT_Q, (long)shutdown);
		}
		while (curproc->wait_cond == (long)shutdown);
	}
	
	/* uk: do an additional sync for the file systems */
	s_ync();
}

# ifdef AUTO_FIX
static char *my_name = "MINT.PRG";

static void get_my_name(void)
{
	register BASEPAGE *bp;
	register DTABUF *dta;
	register char *p;
	
	/* When executing AUTO folder programs, the ROM TOS locates
	 * all programs using Fsfirst/Fsnext; by peeking into our
	 * parent's (the AUTO-execute process's) DTA area, we can
	 * find out our own name...
	 * This works with all known (to me) TOS versions prior to
	 * MultiTOS; but I don't think any other programs should use
	 * this trick, since MiNT or another OS with memory protection
	 * could already be running! <mk>
	 *
	 * Some validity checks first...
	 */
	
	bp = (_base)->p_parent;
	
	if (bp) dta = (DTABUF*)bp->p_dta;
	else dta = NULL;
	
	if (bp && dta)
	{
		p = dta->dta_name ;
		/* Test if "MINT*.PRG" or "MNT*.PRG" */
		if ((strncmp(p, "MINT", 4) == 0 || strncmp(p, "MNT", 3) == 0)
			 && strncmp(p + strlen(p) - 4, ".PRG", 4) == 0)
		{
			my_name = p ;
# if 0
			/* DEBUGGING: */
			boot_print ("[MiNT is named \\AUTO\\%s]\r\n", p);
# endif
		}
	}
}
# endif


/* "boot MiNT?" messages, in various langauges:
 */

static struct yn_message
{
	const char *message;	/* message to print */
	char	yes_let;	/* letter to hit for yes */
	char	no_let;		/* letter to hit for no */
}
boot_it [MAXLANG] =
{
	{ "Load " MINT_NAME "?   (y)es (n)o ", 'y', 'n' },
	{ MINT_NAME " laden?   (j)a (n)ein ", 'j', 'n' },
	{ "Charger " MINT_NAME "?   (o)ui (n)on ", 'o', 'n' },
	{ "Load " MINT_NAME "?   (y)es (n)o ", 'y', 'n' },	/* reserved */
	{ "Cargar " MINT_NAME "?   (s)i (n)o ", 's', 'n' },	/* upside down ? is 168 dec. */
	{ "Carica " MINT_NAME "?   (s)i (n)o ", 's', 'n' }
};

/* ask the user whether s/he wants to boot MultiTOS;
 * returns 1 if yes, 0 if no
 */

static int
boot_kernel_p (void)
{
	extern int gl_lang;
	struct yn_message *msg;
	int y;
	
	msg = &boot_it [gl_lang];
	Cconws (msg->message);
	y = (int) Cconin ();
	if (tolower (y) == msg->yes_let)
		return 1;
	else
		return 0;
}


# ifndef MULTITOS
long GEM_memflags = F_FASTLOAD | F_ALTLOAD | F_ALTALLOC | F_PROT_S | F_ALLOCZERO;
# else
long GEM_memflags = F_FASTLOAD | F_ALTLOAD | F_ALTALLOC | F_PROT_S;
# endif


int
main (int argc, char **argv)
{
	extern int debug_level, debug_logging;	/* in debug.c */
	extern int no_mem_prot;			/* mprot0x0.c */
# ifdef SYSUPDATE_DAEMON
	extern void start_sysupdate();
# endif
	
	static char curpath[128];
	
	long *sysbase;
	long r;
	long yn;
	FILEPTR *f;
	
# ifdef VM_EXTENSION
	int disable_vm = 0;
# endif
	
	/* greetings (placed here 19960610 cpbs to allow me to get version
	 * info by starting MINT.PRG, even if MiNT's already installed.)
	 */
	boot_print (greet1);
	boot_print (greet2);
	
# ifdef __TURBOC__
	boot_print  ("PRELIMINARY PureC compiled version!\r\n");
# endif
	
	/* figure out what kind of machine we're running on
	 * biosfs wants to know this; also sets no_mem_prot
	 * 1992-06-25 kbad put it here so memprot_warning can be intelligent
	 */
	(void) Supexec (getmch);
	
# ifdef ONLY030
# ifndef MMU040
	if (mcpu < 20)
# else
	if (mcpu < 40)
# endif
	{
# ifndef MMU040
		boot_print ("\r\nThis version of MiNT requires a 68020-68060.\r\n");
# else
		boot_print ("\r\nThis version of MiNT requires a 68040-68060.\r\n");
# endif
		boot_print ("Hit any key to continue.\r\n");
		(void) Cconin ();
		Pterm0 ();
	}
# endif
	
	/* Ask the user if s/he wants to boot MiNT */
	if ((Kbshift (-1) & MAGIC_SHIFT) == MAGIC_SHIFT)
	{
		yn = boot_kernel_p ();
		boot_print ("\r\n");
		if (!yn)
			Pterm0 ();
	}
	
	if (argv[0][0] == 0)
	{
		/* maybe started from the desktop */
		curs_off = 1;
	}

	yn = 0;	/* by default, don't print basepage */
	argc--;
	argv++;
	while (argc && **argv == '-')
	{
		if (argv[0][1] >= '0' && argv[0][1] <= '9')
		{
			/* a number sets out_device to that device */
			extern int out_device;
			out_device = (int) atol (&argv[0][1]);
		}
		else if (argv[0][1] == 'b')
		{
			/* print MiNT basepage */
			yn++;
		}
		else if (argv[0][1] == 'd')
		{
			/* -d increases debugging level */
			debug_level++;
		}
# ifdef VM_EXTENSION
		else if (argv[0][1] == 'v')
		{
			/* -v turns off vm */
			disable_vm = 1;
		}
# endif
		else if (argv[0][1] == 'm' || argv[0][1] == 'p')
		{
			/* -m and -p turn off memory protection */
			int givenotice = (argv[0][2] != 'w');
			if (no_mem_prot)
			{
				if (givenotice)
					boot_print (memprot_notice);
			}
			else
			{
				no_mem_prot = 1;
				if (givenotice)
					boot_print (memprot_warning);
			}
		}
		else if (argv[0][1] == 'l')
		{
			/* -l turns on debug logging */
			debug_logging = 1;
		}
		else
		{
			boot_printf ("Unknown argument (ignored): %s\r\n", *argv);
		}
		
		argc--;
		argv++;
	}
	
	if (argc)
	{
		boot_printf ("Unknown argument (ignored): %s (and all the rest)\r\n", *argv);
        }
	
	if (yn)
	{
		boot_printf ("MiNT@%lx\r\nhit a key...", _base);
		(void) Crawcin();
		boot_print ("\r\033K");
	}
	
# if 0
	/* if less than 1 megabyte free, turn off memory protection */
	if (Mxalloc (-1L, 3) < ONE_MEG && !no_mem_prot)
	{
		extern const char *insuff_mem_warning;
		boot_print (insuff_mem_warning);
		no_mem_prot = 1;
	}
# endif
	
# ifndef AUTO_FIX
	/* look for ourselves as \AUTO\MINTNP.PRG; if so, we turn memory
	 * protection off
	 */
	if (!no_mem_prot && Fsfirst("\\AUTO\\MINTNP.PRG", 0) == 0)
	{
		no_mem_prot = 1;
	}
# endif
	
# ifdef OLDTOSFS
	/* Get GEMDOS version from ROM for later use by our own Sversion() */
	gemdos_version = Sversion ();
# endif	
	
	/* check for GEM -- this must be done from user mode */
	gem_active = check_for_gem ();
	
# ifdef AUTO_FIX
	/* only useful if we're in the AUTO folder... */
	if (!gem_active)
	{
		get_my_name ();
		
# ifdef VM_EXTENSION		
		/* turn off vm and mem prot if name
		 * matches "MINTNM*.*" or "MNTNM*.*"
		 */
		if (!disable_vm)
		{
			if (strncmp (my_name, "MINTNM", 6) == 0
				|| strncmp (my_name, "MNTNM", 5) == 0)
			{
				disable_vm = 1;
				no_mem_prot = 1;
			}
		}
# endif		
		/* Turn off memory protection if our name matches either
		 * "MINTN*.PRG" or "MNTN*.PRG":
		 * (will enable vm)
		 */
		if (!no_mem_prot)
		{
				if (strncmp(my_name, "MINTN", 5) == 0
					|| strncmp(my_name, "MNTN", 4) == 0)
				{
					no_mem_prot = 1;
				}
		}
	}
# endif
# ifdef VERBOSE_BOOT
	boot_print ("Memory protection ");
	if (no_mem_prot)
		boot_print ("disabled\r\n");
	else
		boot_print ("enabled\r\n");
# endif
	
	/* get the current directory, so that we can switch back to it after
	 * the file systems are properly initialized
	 *
	 * set the current directory for the current process
	 */
	Dgetpath (curpath, 0);
	if (!*curpath)
	{
		curpath[0] = '\\';
		curpath[1] = 0;
	}
	tosssp = (long) Super (0L);	/* enter supervisor mode */
	
	if (!no_mem_prot)
		save_mmu ();		/* save current MMU setup */
	
	/* get GEMDOS pointer to current basepage
	 * 
	 * 0x4f2 points to the base of the OS; here we can find the OS compilation
	 * date, and (in newer versions of TOS) where the current basepage pointer
	 * is kept; in older versions of TOS, it's at 0x602c
	 */
	sysbase = *((long **)(0x4f2L));	/* gets the RAM OS header */
	sysbase = (long *)sysbase[2];	/* gets the ROM one */

	tosvers = (int)(sysbase[0] & 0x0000ffff);
	falcontos = (tosvers >= 0x0400 && tosvers <= 0x0404) || (tosvers >= 0x0700);
	if (tosvers == 0x100)
	{
		if ((sysbase[7] & 0xfffe0000L) == 0x00080000L)
			tosbp = (BASEPAGE **)0x873cL;	/* SPANISH ROM */
		else
			tosbp = (BASEPAGE **) 0x602cL;
		kbshft = (char *) 0x0e1bL;
	}
	else
	{
		tosbp = (BASEPAGE **) sysbase[10];
		kbshft = (char *) sysbase[9];
	}

	if (falcontos)
	{
		bconmap2 = (BCONMAP2_T *) Bconmap (-2);
		if (bconmap2->maptabsize == 1)
		{
			/* Falcon BIOS Bconmap is busted */
			bconmap2->maptabsize = 3;
		}
		has_bconmap = 1;
	}
	else
	{
		/* The TT TOS release notes are wrong;
		 * this is the real way to test for Bconmap ability
		 */
		has_bconmap = (Bconmap (0) == 0);
		if (has_bconmap)
			bconmap2 = (BCONMAP2_T *) Bconmap (-2);
	}
	
# ifdef VM_EXTENSION
	if ((mcpu == 30) && (no_mem_prot) && (!disable_vm))
	{
		/* use vm */
		vm_in_use = 1;
	}
# endif
	
	/* init XBIOS Random() function */
	init_random();
	DEBUG (("%s, %ld: init_rnd() ok!", __FILE__, (long) __LINE__));
 
	/* initialize cache */
	init_cache ();
	DEBUG (("%s, %ld: init_cache() ok!", __FILE__, (long) __LINE__));
	
	/* initialize memory */
	init_mem ();
	DEBUG (("%s, %ld: init_mem() ok!", __FILE__, (long) __LINE__));
	
	/* Initialize high-resolution calendar time */
	init_time ();
	DEBUG (("%s, %ld: init_time() ok!", __FILE__, (long) __LINE__));
	
	/* initialize buffered block I/O */
	init_block_IO ();
	DEBUG (("%s, %ld: init_block_IO() ok!", __FILE__, (long) __LINE__));
	
	/* initialize the basic file systems */
	init_filesys ();
	DEBUG (("%s, %ld: init_filesys() ok!", __FILE__, (long) __LINE__));
	
	/* initialize processes */
	init_proc();
	DEBUG (("%s, %ld: init_proc() ok!", __FILE__, (long) __LINE__));
	
	/* initialize system calls */
	init_dos ();
	DEBUG (("%s, %ld: init_dos() ok!", __FILE__, (long) __LINE__));
	init_bios ();
	DEBUG (("%s, %ld: init_bios() ok!", __FILE__, (long) __LINE__));
	init_xbios ();
	DEBUG (("%s, %ld: init_xbios() ok!", __FILE__, (long) __LINE__));
	
	/* NOTE: there's a call to kmalloc embedded in install_cookies, which
	 * is called by init_intr; so make sure this is the last of the
	 * init_* things called!
	 */
	init_intr ();
	DEBUG (("%s, %ld: init_intr() ok!", __FILE__, (long) __LINE__));
	
	/* add our pseudodrives */
	*((long *) 0x4c2L) |= PSEUDODRVS;
	
	/* we'll be making GEMDOS calls */
	enter_gemdos ();
	
# if 0
	if (!gem_active)
	{
		/* make MiNT invisible in the basepage chain, so that
		 * programs that rely on a certain basepage chain
		 * structure to determine whether or not they were run
		 * from the desktop will have a better chance of working.
		 * NOTE THAT THIS IS ONLY DONE TO HELP OUT BRAIN-DAMAGED
		 * SOFTWARE: do *not* try counting basepages to figure
		 * out whether or not you were run from the desktop!!!
		 */
		rootproc->base = _base->p_parent;
	}
	else
# endif
		rootproc->base = _base;
	
	/* set up standard file handles for the current process
	 * do this here, *after* init_intr has set the Rwabs vector,
	 * so that AHDI doesn't get upset by references to drive U:
	 */
	f = do_open ("U:\\DEV\\CONSOLE", O_RDWR, 0, NULL);
	if (!f)
	{
		FATAL ("unable to open CONSOLE device");
	}
	curproc->control = f;
	curproc->handle[0] = f;
	curproc->handle[1] = f;
	f->links = 3;

	f = do_open ("U:\\DEV\\MODEM1", O_RDWR, 0, NULL);
	curproc->aux = f;
	((struct tty *) f->devinfo)->aux_cnt = 1;
	f->pos = 1;	/* flag for close to --aux_cnt */
	
	if (has_bconmap)
	{
		/* If someone has already done a Bconmap call, then
		 * MODEM1 may no longer be the default
		 */
		bconmap (curbconmap);
		f = curproc->aux;	/* bconmap can change curproc->aux */
	}
	
	if (f)
	{
		curproc->handle[2] = f;
		f->links++;
	}
	
	f = do_open ("U:\\DEV\\CENTR", O_RDWR, 0, NULL);
	if (f)
	{
		curproc->handle[3] = curproc->prn = f;
		f->links = 2;
	}
	
	f = do_open ("U:\\DEV\\MIDI", O_RDWR, 0, NULL);
	if (f)
	{
		curproc->midiin = curproc->midiout = f;
		((struct tty *)f->devinfo)->aux_cnt = 1;
		f->pos = 1;	/* flag for close to --aux_cnt */
		f->links = 2;
	}
	
	/* load external file systems
	 * 
	 * set path first to make sure that MiNT's directory matches
	 * GEMDOS's
	 */
	d_setpath (curpath);
	
	load_devdriver ();
	
# ifndef PROFILING
	/* load_filesys causes media changes :-( */
	load_filesys ();
# endif

	/* note that load_filesys changed the
	 * directory on us!!
	 */
	d_setpath (curpath);
	
	/* load the configuration file */
	load_config ();
	
	/* start system update daemon */
	if (sync_time > 0)
	{
# ifdef SYSUPDATE_DAEMON
		start_sysupdate();
# else
		addroottimeout (1000L * sync_time,
			(void _cdecl (*)(PROC *p)) do_sync, 0);
# endif
	}
	
	if (init_env == 0)
		init_env = (char *) _base->p_env;

	/* empty environment?
	 * Set the PATH variable to the root of the current drive
	 */
	if (init_env[0] == 0)
	{
		static char path_env[] = "PATH=\0C:\0";
		path_env[6] = curproc->curdrv + 'A';
		init_env = path_env;
	}

	/* if we are MultiTOS, we're running in the AUTO folder, and our INIT
	 * is in fact GEM, take the exec_os() vector. (We know that INIT
	 * is GEM if the user told us so by using GEM= instead of INIT=.)
	 */
	if (!gem_active && init_is_gem)
	{
		xbra_install (&old_execos, EXEC_OS, (long _cdecl (*)())do_exec_os);
	}
	
	/* run any programs appearing after us in the AUTO folder */
	run_auto_prgs ();
	
	/* prepare to run the init program as PID 1. */
	set_pid_1 ();
	
	/* run the initial program
	 *
	 * if that program is in fact GEM, we start it via exec_os, otherwise
	 * we do it with Pexec.
	 * the logic is: if the user specified init_prg, and it is not
	 * GEM, then we try to execute it; if it *is* GEM (e.g. gem.sys),
	 * then we try to execute it if gem is already active, otherwise
	 * we jump through the exec_os vector (which we grabbed above) in
	 * order to start it. We *never* go through exec_os if we're not in
	 * the AUTO folder.
	 */
	if (init_prg && (!init_is_gem || gem_active))
	{
		r = p_exec (0, (char *) init_prg, init_tail, init_env);
	}
	else if (!gem_active)
	{   
		BASEPAGE *bp; int pid;
		bp = (BASEPAGE *) p_exec (7, (char *) GEM_memflags, (char *) "\0", init_env);
		bp->p_tbase = *((long *) EXEC_OS);
# ifndef MULTITOS
		if (((long *) sysbase[5])[0] == 0x87654321)
			gem_start = ((long *) sysbase[5])[2];
		gem_base = bp;
# endif
		r = p_exec (106, (char *) "GEM", bp, 0L);
		pid = (int) r;
		if (pid > 0)
		{
			do {
				r = p_wait3(0, (long *)0);
			}
			while (pid != ((r & 0xffff0000L) >> 16));
			r &= 0x0000ffff;
		}
	}
	else
	{
		boot_print ("If MiNT is run after GEM starts, you must specify a program\r\n");
		boot_print ("to run initially in MINT.CNF, with an INIT= line\r\n");
		r = 0;
	}
	
	if (r < 0 && init_prg)
	{
		boot_printf ("FATAL: couldn't run %s\r\n", init_prg);
	}
	
	if (r)
	{
		boot_printf ("exit code: %ld\r\n", r);
	}
	
	rootproc->base = _base;

	/* shut down all processes gracefully */
	shutdown();

	/* put everything back and exit */
	if (!gem_active && init_is_gem)
	{
		/* we stole exec_os above */
		*((long *)EXEC_OS) = (long)old_execos.next;
	}
	
	restr_intr();
	close_filesys();
	if (!no_mem_prot) restr_mmu();
	restr_screen();
	
	(void) Super (tosssp);
	
	boot_print ("leaving MiNT\r\n");
	
	if (curs_off) boot_print ("\033f");	/* disable cursor */
	
	return 0;
}


/*
 * cookie jar handling routines. The "cookie jar" is an area of memory
 * reserved by TOS for TSR's and utility programs; the idea is that
 * you put a cookie in the jar to notify people of available services.
 * The BIOS uses the cookie jar in TOS 1.6 and higher; for earlier versions
 * of TOS, the jar is always empty (unless someone added a cookie before
 * us; POOLFIX does, for example).
 * MiNT establishes an entirely new cookie jar (with the old cookies copied
 * over) and frees it on exit. That's because TSR's run under MiNT
 * will no longer be accessible after MiNT exits.
 * MiNT also puts a cookie in the jar, with tag field 'MiNT' (of course)
 * and with the major version of MiNT in the high byte of the low word,
 * and the minor version in the low byte.
 */

void
install_cookies (void)
{
	COOKIE *cookie;
	ushort i = 0, ncookies = 0;

	long ncsize;
	extern long rsvf;

	/* note that init_intr sets oldcookie to the old cookie jar */

	cookie = oldcookie;
	if (cookie)
	{
		while (cookie->tag != 0)
		{
			/* check for _FLK cookie */
			if (cookie->tag == COOKIE__FLK)
				flk = 1;
		
			/* ..and for RSVF */
			else if (cookie->tag == COOKIE_RSVF)
				rsvf = cookie->value;

# if !defined(CPU_TEST) && !defined(FPU_TEST)
			/* check for true FPU co-processor */
			else if ((cookie->tag == COOKIE__FPU) \
					&& (cookie->value >> 16) >= 2)
				fpu = 1;
# endif
			cookie++;
			ncookies++;
		}
	}
	
	ncsize = MIN (cookie->value, 240);	/* avoid to big tag values */
	if (ncookies > ncsize)
		ncsize = ncookies;
	
	/*
	 * We allocate the cookie jar in global memory so anybody can read
	 * it or write it. This code allocates at least 16 more cookies,
	 * then rounds up to a QUANTUM boundary (that's what ROUND does). 
	 * Probably, nobody will have to allocate another cookie jar :-)
	 *
	 * NOTE: obviously, we can do this only if init_intr is called
	 * _after_ memory, processes, etc. have been initialized
	 */
	ncsize = (ncsize + 16) * sizeof (COOKIE);
	ncsize = ROUND (ncsize);
	newjar_region = get_region (core, ncsize, PROT_G);
	newcookie = (COOKIE *) attach_region (rootproc, newjar_region);
	
# ifdef CPU_TEST
	/* set the hardware detected CPU and FPU rather
	 * than trust the TOS
	 */
	newcookie[i].tag = COOKIE__CPU;
	newcookie[i].value = mcpu; 
	i++;
# endif
# if defined(CPU_TEST) && defined(FPU_TEST)
	if (fputype) {
		newcookie[i].tag = COOKIE__FPU;
		newcookie[i].value = fputype;
		i++;
	}
# endif
	/* copy the old cookies to the new jar */
	cookie = oldcookie;

	while (ncookies) {
		/* but don't copy RSVF, MiNTs /dev is for real...
		 * (if you want to know whats in there use ls :)
		 */
# ifndef CPU_TEST
		if (cookie->tag == COOKIE_RSVF)
			cookie++;
# else
# ifndef FPU_TEST
		/* we already installed own _CPU and _FPU cookies */
		if (cookie->tag == COOKIE_RSVF || cookie->tag == COOKIE__CPU)
			cookie++;
# else
		if (cookie->tag == COOKIE_RSVF \
				|| cookie->tag == COOKIE__CPU \
					|| cookie->tag == COOKIE__FPU)
			cookie++;
# endif
# endif
		else
			newcookie[i++] = *cookie++;
		ncookies--;
	}
	
	/* install MiNT cookie */
	newcookie[i].tag   = COOKIE_MiNT;
	newcookie[i].value = (MAJ_VERSION << 8) | MIN_VERSION;
	i++;
	
	/* install _FLK cookie to indicate that file locking works */
	if (!flk)
	{
		newcookie[i].tag   = COOKIE__FLK;
		newcookie[i].value = 0;
		i++;
	}
	
	/* jr: install PMMU cookie if memory protection is used
	 * Draco: also do the same if VM is active; memory protection
	 * and VM exclude each other, but some day we can get both
	 * working simultaneously, who knows...
	 */
# ifdef	VM_EXTENSION
	if (!no_mem_prot || vm_in_use)
	{
# else
	if (!no_mem_prot)
	{
# endif
		newcookie[i].tag   = COOKIE_PMMU;
		newcookie[i].value = 0;
		i++;
	}
	
	/* the last cookie should have a 0 tag, and a value indicating
	 * the number of slots, total
	 */
	newcookie[i].tag   = 0;
	newcookie[i].value = ncsize / sizeof(COOKIE);
	
	*CJAR = newcookie;
}

/*
 * Get the value of the _MCH cookie, if one exists; also set no_mem_prot if
 * there's a _CPU cookie and you're not on an '030, or if there is none.
 * This must be done in a separate routine because the machine type and CPU
 * type are needed when initializing the system, whereas install_cookies is
 * not called until everything is practically up.
 * In fact, getmch() should be called before *anything* else is
 * initialized, so that if we find a MiNT cookie already in the
 * jar we can bail out early and painlessly.
 */

static long
getmch (void)
{
	extern int no_mem_prot;
	
	COOKIE *jar = *CJAR;	/* CJAR defined in cookie.h */

# ifndef CPU_TEST
	int foundcpu = 0;
	
	mcpu = 0;		/* get info from TOS */
# else
	mcpu = detect_cpu();	/* own hardware test */
# endif
# ifdef FPU_TEST

	fputype = detect_fpu();	/* this must be done after the CPU detection */
	if ((fputype >> 16) > 1)
		fpu = 1;

# endif
	
	DEBUG (("detecting hardware ... "));
	/* at the moment only detection of ST-ESCC */
	if (detect_hardware ())
		boot_print ("ST-ESCC extension detected\r\n");
	DEBUG (("ok!\r\n"));
	
	if (jar)
	{
		while (jar->tag != 0)
		{
			/* check for machine type */
			if (jar->tag == COOKIE__MCH)
			{
				mch = jar->value;
# ifdef MILAN
				if (mch != MILAN_C)
				{
					boot_print ("This MiNT version requires a Milan!!!\r\n");
					Pterm(2);
				}
# else
				if (mch == MILAN_C)
				{
					boot_print ("This MiNT version doesn't run on a Milan!!!\r\n");
					Pterm(2);
				}
# endif
			}
# ifndef CPU_TEST
			else if (jar->tag == COOKIE__CPU)
			{
	    			/* if not '030 then no memory protection */
				mcpu = jar->value;
# ifndef MMU040
	    			if (mcpu != 30) no_mem_prot = 1;
# endif
	    			foundcpu = 1;
			}
# endif
			else if (jar->tag == COOKIE__VDO)
			{
				FalconVideo = (jar->value == 0x00030000L);
				ste_video = (jar->value == 0x00010000L);
				if (jar->value & 0xffff0000L)
					screen_boundary = 15;
			}
			else if (jar->tag == COOKIE_MiNT)
			{
				boot_print ("MiNT is already installed!!\r\n");
				Pterm(2);
			}
			else if (jar->tag == COOKIE__AKP)
			{
				gl_lang = (int) ((jar->value >> 8) & 0x00ff);
			}
			else if (jar->tag == COOKIE_PMMU)
			{
				/* jr: if PMMU cookie exists, someone else is
				 * already using the PMMU
				 */
				boot_print ("Warning: PMMU is already in use\r\n");
				no_mem_prot = 1;
# ifdef VM_EXTENSION
				vm_in_use = 0;
# endif
			}
			jar++;
		}
	}

# ifndef CPU_TEST	
	if (!foundcpu) no_mem_prot = 1;
# elif !defined(MMU040)
	if (mcpu != 30) no_mem_prot = 1;
# endif
	
	/*
	 * if no preference found, look at the country code to decide
	 */
	if (gl_lang < 0)
	{
		long *sysbase;
		int i;
		
		sysbase = *((long **)(0x4f2L)); /* gets the RAM OS header */
		sysbase = (long *)sysbase[2];	/* gets the ROM one */
		
		i = (int) ((sysbase[7] & 0x7ffe0000L) >> 17L);
		
		switch (i)
		{
			case 1:		/* Germany */
			case 8:		/* Swiss German */
				gl_lang = 1;
				break;
			case 2:		/* France */
			case 7:		/* Swiss French */
				gl_lang = 2;
				break;
			case 4:		/* Spain */
				gl_lang = 4;
				break;
			case 5:		/* Italy */
				gl_lang = 5;
				break;
			default:
				gl_lang = 0;
				break;
		}
	}
	
	if (gl_lang >= MAXLANG || gl_lang < 0)
		gl_lang = 0;
	
	return 0L;
}

/*
 * run programs in the AUTO folder that appear after MINT.PRG
 * some things to watch out for:
 * (1) make sure GEM isn't active
 * (2) make sure there really is a MINT.PRG in the auto folder
 */

/*
 * some global variables used to see if GEM is active
 */

static short aes_intout[64];
static short aes_dummy[64];
static short aes_globl[15];
static short aes_cntrl[6] = { 10, 0, 1, 0, 0 };

short *aes_pb[6] = { aes_cntrl, aes_globl, aes_dummy, aes_intout,
		     aes_dummy, aes_dummy };

/*
 * check for whether GEM is active; remember, this *must* be done in
 * user mode
 */

static int
check_for_gem (void)
{
	call_aes(aes_pb);	/* does an appl_init */
	return aes_globl[0];
}

static void
run_auto_prgs (void)
{
	DTABUF *dta;
	long r;
	static char pathspec[32] = "\\AUTO\\";
	short runthem = 0;	/* set to 1 after we find MINT.PRG */
	char curpath[PATH_MAX];
 	int curdriv, bootdriv;

	/* if the AES is running, don't check AUTO */
	if (gem_active)	return;

	/* OK, now let's run through \\AUTO looking for
	 * programs...
	 */
	d_getpath(curpath,0);
	curdriv = d_getdrv();
	
	/* We are in Supervisor mode, so we can do this */
	bootdriv=*((short*)0x446);
	d_setdrv(bootdriv);
	d_setpath("\\");

	dta = (DTABUF *)f_getdta();
	r = f_sfirst("\\AUTO\\*.PRG", 0);
	while (r >= 0)
	{
# ifdef AUTO_FIX
		if (strcmp(dta->dta_name, my_name) == 0)
# else
		if (!strcmp(dta->dta_name, "MINT.PRG")
			|| !strcmp(dta->dta_name, "MINTNP.PRG" ))
# endif
		{
			runthem = 1;
		}
		else if (runthem)
		{
			strcpy(pathspec+6, dta->dta_name);
			p_exec(0, pathspec, (char *)"", init_env);
		}
		r = f_snext();
	}
	
 	d_setdrv(curdriv);
 	d_setpath(curpath);
}
