/*
 * This file has been modified as part of the FreeMiNT project. See
 * the file Changes.MH for details and dates.
 */

/*
 * Copyright 1992,1993 Eric R. Smith.
 * Copyright 1993,1994 Atari Corporation.
 * All rights reserved.
 */

/* Shared memory file system */

# include "shmfs.h"
# include "global.h"

# include "biosfs.h"
# include "memory.h"
# include "kmemory.h"
# include "string.h"


static long	_cdecl shm_root		P_((int drv, fcookie *fc));
static long	_cdecl shm_creat	P_((fcookie *dir, const char *name, unsigned mode, int attrib, fcookie *fc));
static long	_cdecl shm_lookup	P_((fcookie *dir, const char *name, fcookie *fc));
static long	_cdecl shm_getxattr	P_((fcookie *fc, XATTR *xattr));
static long	_cdecl shm_chattr	P_((fcookie *fc, int attrib));
static long	_cdecl shm_chown	P_((fcookie *fc, int uid, int gid));
static long	_cdecl shm_chmode	P_((fcookie *fc, unsigned mode));
static long	_cdecl shm_rmdir	P_((fcookie *dir, const char *name));
static long	_cdecl shm_remove	P_((fcookie *dir, const char *name));
static long	_cdecl shm_getname	P_((fcookie *root, fcookie *dir, char *pathname, int size));
static long	_cdecl shm_rename	P_((fcookie *olddir, char *oldname, fcookie *newdir, const char *newname));
static long	_cdecl shm_opendir	P_((DIR *dirh, int flags));
static long	_cdecl shm_readdir	P_((DIR *dirh, char *nm, int nmlen, fcookie *));
static long	_cdecl shm_rewinddir	P_((DIR *dirh));
static long	_cdecl shm_closedir	P_((DIR *dirh));
static long	_cdecl shm_pathconf	P_((fcookie *dir, int which));
static long	_cdecl shm_dfree	P_((fcookie *dir, long *buf));
static long	_cdecl shm_fscntl 	P_((fcookie *dir, const char *name, int cmd, long arg));
static DEVDRV *	_cdecl shm_getdev	P_((fcookie *fc, long *devsp));

static long	_cdecl shm_open		P_((FILEPTR *f));
static long	_cdecl shm_write	P_((FILEPTR *f, const char *buf, long bytes));
static long	_cdecl shm_read		P_((FILEPTR *f, char *buf, long bytes));
static long	_cdecl shm_lseek	P_((FILEPTR *f, long where, int whence));
static long	_cdecl shm_ioctl	P_((FILEPTR *f, int mode, void *buf));
static long	_cdecl shm_datime	P_((FILEPTR *f, ushort *time, int rwflag));
static long	_cdecl shm_close	P_((FILEPTR *f, int pid));

/* dummy routines from biosfs.c */
extern long	_cdecl null_select	P_((FILEPTR *f, long p, int mode));
extern void	_cdecl null_unselect	P_((FILEPTR *f, long p, int mode));

static ushort shmtime, shmdate;

#define SHMNAME_MAX 15

typedef struct shmfile
{
	struct shmfile *next;
	char filename[SHMNAME_MAX+1];
	int uid, gid;
	ushort time, date;
	unsigned mode;
	int inuse;
	MEMREGION *reg;
} SHMFILE;

SHMFILE *shmroot = 0;

DEVDRV shm_device =
{
	shm_open, shm_write, shm_read, shm_lseek, shm_ioctl, shm_datime,
	shm_close, null_select, null_unselect
};

FILESYS shm_filesys =
{
	NULL,
	FS_LONGPATH | FS_NO_C_CACHE,
	shm_root,
	shm_lookup, shm_creat, shm_getdev, shm_getxattr,
	shm_chattr, shm_chown, shm_chmode,
	nomkdir, shm_rmdir, shm_remove, shm_getname, shm_rename,
	shm_opendir, shm_readdir, shm_rewinddir, shm_closedir,
	shm_pathconf, shm_dfree,
	nowritelabel, noreadlabel, nosymlink, noreadlink, nohardlink,
	shm_fscntl, nodskchng
};

long _cdecl 
shm_root(int drv, fcookie *fc)
{
	if ((unsigned)drv == SHMDRV) {
		fc->fs = &shm_filesys;
		fc->dev = drv;
		fc->index = 0L;
		return E_OK;
	}
	fc->fs = 0;
	return EINTRN;
}

static long _cdecl 
shm_lookup(fcookie *dir, const char *name, fcookie *fc)
{
	SHMFILE *s;

	if (dir->index != 0) {
		DEBUG(("shm_lookup: bad directory"));
		return EPTHNF;
	}

/* special case: an empty name in a directory means that directory */
/* so does "." */
	if (!*name || (name[0] == '.' && name[1] == 0)) {
		*fc = *dir;
		return E_OK;
	}

/* another special case: ".." could be a mount point */
	if (!strcmp(name, "..")) {
		*fc = *dir;
		return EMOUNT;
	}

	for (s = shmroot; s; s = s->next) {
		if (!stricmp(s->filename,name))
			break;
	}

	if (!s) {
		DEBUG(("shm_lookup: name not found"));
		return EFILNF;
	} else {
		fc->index = (long)s;
		fc->fs = &shm_filesys;
		fc->dev = SHMDRV;
	}
	return E_OK;
}

static long _cdecl 
shm_getxattr(fcookie *fc, XATTR *xattr)
{
	SHMFILE *s;

	xattr->blksize = 1;
	if (fc->index == 0) {
		/* the root directory */
		xattr->index = 0;
		xattr->dev = xattr->rdev = SHMDRV;
		xattr->nlink = 1;
		xattr->uid = xattr->gid = 0;
		xattr->size = xattr->nblocks = 0;
		xattr->mtime = xattr->atime = xattr->ctime = shmtime;
		xattr->mdate = xattr->adate = xattr->cdate = shmdate;
		xattr->mode = S_IFDIR | DEFAULT_DIRMODE;
		xattr->attr = FA_DIR;
		return E_OK;
	}

	s = (SHMFILE *)fc->index;
	xattr->index = (long) s;
	xattr->dev = SHMDRV;
	xattr->rdev = PROC_RDEV_BASE | 0;
	xattr->uid = s->uid; xattr->gid = s->gid;
	if (s->reg) {
		xattr->size = xattr->nblocks = s->reg->len;
		xattr->nlink = s->reg->links + 1;
	 } else {
		xattr->size = xattr->nblocks = 0;
		xattr->nlink = 1;
	}
	xattr->mtime = xattr->ctime = xattr->atime = s->time;
	xattr->mdate = xattr->cdate = xattr->adate = s->date;
	xattr->mode = s->mode;
	xattr->attr = (s->mode & (S_IWUSR|S_IWGRP|S_IWOTH)) ? 0 : FA_RDONLY;
	return E_OK;
}

static long _cdecl 
shm_chattr(fcookie *fc, int attrib)
{
	SHMFILE *s;

	s = (SHMFILE *)fc->index;
	if (!s) return EACCDN;

	if (attrib & FA_RDONLY) {
		s->mode &= ~(S_IWUSR|S_IWGRP|S_IWOTH);
	} else if ( !(s->mode & (S_IWUSR|S_IWGRP|S_IWOTH)) ) {
		s->mode |= (S_IWUSR|S_IWGRP|S_IWOTH);
	}
	return E_OK;
}

static long _cdecl 
shm_chown(fcookie *fc, int uid, int gid)
{
	SHMFILE *s;

	s = (SHMFILE *)fc->index;
	if (!s)
		return EACCDN;
	s->uid = uid;
	s->gid = gid;
	return E_OK;
}

static long _cdecl 
shm_chmode(fcookie *fc, unsigned int mode)
{
	SHMFILE *s;

	s = (SHMFILE *)fc->index;
	if (!s)
		return EINVFN;
	s->mode = mode;
	return E_OK;
}

static long _cdecl 
shm_rmdir(fcookie *dir, const char *name)
{
	UNUSED(dir); UNUSED(name);

	return EPTHNF;
}

static long _cdecl 
shm_remove(fcookie *dir, const char *name)
{
	SHMFILE *s, **old;

	if (dir->index != 0)
		return EPTHNF;

	old = &shmroot;
	for (s = shmroot; s; s = s->next) {
		if (!stricmp(s->filename, name))
			break;
		old = &s->next;
	}
	if (!s)
		return EFILNF;
	if (s->inuse)
		return EACCDN;
	*old = s->next;

	s->reg->links--;
	if (s->reg->links <= 0) {
		free_region(s->reg);
	}
	kfree(s);
	shmtime = timestamp;
	shmdate = datestamp;
	return E_OK;
}

static long _cdecl 
shm_getname(fcookie *root, fcookie *dir, char *pathname, int size)
{
	UNUSED(root); UNUSED(dir);

/* BUG: 'size' should be used in a more meaningful way */
	if (size <= 0) return ENAMETOOLONG;
	*pathname = 0;
	return E_OK;
}

static long _cdecl 
shm_rename(fcookie *olddir, char *oldname, fcookie *newdir, const char *newname)
{
	SHMFILE *s;

	if (olddir->index != 0 || newdir->index != 0)
		return EPTHNF;

/* verify that "newname" doesn't exist */
	for (s = shmroot; s; s = s->next)
		if (!strnicmp(s->filename, newname, SHMNAME_MAX))
			return EACCDN;

	for (s = shmroot; s; s = s->next)
		if (!strnicmp(s->filename, oldname, SHMNAME_MAX))
			break;
	if (!s)
		return EFILNF;

	strncpy(s->filename, newname, SHMNAME_MAX);
	shmtime = timestamp;
	shmdate = datestamp;
	return E_OK;
}

static long _cdecl 
shm_opendir(DIR *dirh, int flags)
{
	UNUSED(flags);

	dirh->index = 0;
	return E_OK;
}

static long _cdecl 
shm_readdir(DIR *dirh, char *name, int namelen, fcookie *fc)
{
	int i;
	int giveindex = (dirh->flags == 0);
	SHMFILE *s;

	s = shmroot;
	i = dirh->index++;
	while (i > 0 && s != 0) {
		s = s->next;
		--i;
	}
	if (!s)
		return ENMFIL;

	fc->index = (long)s;
	fc->fs = &shm_filesys;
	fc->dev = SHMDRV;

	if (giveindex) {
		namelen -= (int)sizeof(long);
		if (namelen <= 0)
			return ERANGE;
		*((long *)name) = (long)s;
		name += sizeof(long);
	}
	if (namelen <= strlen(s->filename))
		return ENAMETOOLONG;
	strcpy(name, s->filename);
	return E_OK;
}

static long _cdecl 
shm_rewinddir(DIR *dirh)
{
	dirh->index = 0;
	return E_OK;
}

static long _cdecl 
shm_closedir(DIR *dirh)
{
	UNUSED(dirh);

	return E_OK;
}

static long _cdecl 
shm_pathconf(fcookie *dir, int which)
{
	UNUSED(dir);

	switch(which) {
	case DP_INQUIRE:
		return DP_XATTRFIELDS;
	case DP_IOPEN:
		return UNLIMITED;	/* no internal limit on open files */
	case DP_MAXLINKS:
		return 1;		/* we don't have hard links */
	case DP_PATHMAX:
		return PATH_MAX;	/* max. path length */
	case DP_NAMEMAX:
		return SHMNAME_MAX;	/* max. length of individual name */
	case DP_ATOMIC:
		return UNLIMITED;	/* all writes are atomic */
	case DP_TRUNC:
		return DP_AUTOTRUNC;	/* file names are truncated */
	case DP_CASE:
		return DP_CASEINSENS;	/* case preserved, but ignored */
	case DP_MODEATTR:
		return (0777L << 8)|
				DP_FT_DIR|DP_FT_MEM;
	case DP_XATTRFIELDS:
		return DP_INDEX|DP_DEV|DP_NLINK|DP_UID|DP_GID|DP_BLKSIZE|DP_SIZE|
				DP_NBLOCKS|DP_MTIME;
	default:
		return EINVFN;
	}
}

static long _cdecl 
shm_dfree(fcookie *dir, long int *buf)
{
	long size;
/* "sector" size is the size of the smallest amount of memory that can be
   allocated. see mem.h for the definition of ROUND
 */
	long secsiz = ROUND(1);

	UNUSED(dir);

	size = tot_rsize(core, 0) + tot_rsize(alt, 0);
	*buf++ = size/secsiz;			/* number of free clusters */
	size = tot_rsize(core, 1) + tot_rsize(alt, 1);
	*buf++ = size/secsiz;			/* total number of clusters */
	*buf++ = secsiz;			/* sector size (bytes) */
	*buf = 1;				/* cluster size (in sectors) */
	return E_OK;
}

static long _cdecl
shm_fscntl (fcookie *dir, const char *name, int cmd, long arg)
{	
	UNUSED(dir); UNUSED(name);

	switch (cmd)
	{
		case MX_KER_XFSNAME:
		{
			strcpy ((char *) arg, "shm-xfs");
			return E_OK;
		}
	}
	return EINVFN;
}

static DEVDRV * _cdecl 
shm_getdev(fcookie *fc, long int *devsp)
{
	SHMFILE *s;

	s = (SHMFILE *)fc->index;

	*devsp = (long)s;
	return &shm_device;
}

/*
 * create a shared memory region
 */

static long _cdecl 
shm_creat(fcookie *dir, const char *name, unsigned int mode, int attrib, fcookie *fc)
{
	SHMFILE *s;

	UNUSED(attrib);
/*
 * see if the name already exists
 */
	for (s = shmroot; s; s = s->next) {
		if (!stricmp(s->filename, name)) {
			DEBUG(("shm_creat: file exists"));
			return EACCDN;
		}
	}
	
	s = kmalloc (sizeof (*s));
	if (!s)
		return ENSMEM;

	s->inuse = 0;
	strncpy(s->filename, name, SHMNAME_MAX);
	s->filename[SHMNAME_MAX] = 0;
	s->uid = curproc->euid;
	s->gid = curproc->egid;
	s->mode = mode;
	s->next = shmroot;
	s->reg = 0;
	s->time = shmtime = timestamp;
	s->date = shmdate = datestamp;
	shmroot = s;

	fc->fs = &shm_filesys;
	fc->index = (long)s;
	fc->dev = dir->dev;

	return E_OK;
}

/*
 * Shared memory device driver
 */

/*
 * BUG: file locking and the O_SHMODE restrictions are not implemented
 * for shared memory
 */

static long _cdecl 
shm_open(FILEPTR *f)
{
	SHMFILE *s;

	s = (SHMFILE *)f->devinfo;
	s->inuse++;
	return E_OK;
}

static long _cdecl 
shm_write(FILEPTR *f, const char *buf, long int nbytes)
{
	SHMFILE *s;
	char *where;
	long bytes_written = 0;

	s = (SHMFILE *)f->devinfo;
	if (!s->reg)
		return 0;

	if (nbytes + f->pos > s->reg->len)
		nbytes = s->reg->len - f->pos;

	where = (char *)s->reg->loc + f->pos;

/* BUG: memory read/writes should check for valid addresses */

TRACE(("shm_write: %ld bytes to %lx", nbytes, where));

	while (nbytes-- > 0) {
		*where++ = *buf++;
		bytes_written++;
	}
	f->pos += bytes_written;
	s->time = timestamp;
	s->date = datestamp;
	return bytes_written;
}

static long _cdecl 
shm_read(FILEPTR *f, char *buf, long int nbytes)
{
	SHMFILE *s;
	char *where;
	long bytes_read = 0;

	s = (SHMFILE *)f->devinfo;
	if (!(s->reg))
		return 0;

	if (nbytes + f->pos > s->reg->len)
		nbytes = s->reg->len - f->pos;

	where = (char *)s->reg->loc + f->pos;

TRACE(("shm_read: %ld bytes from %lx", nbytes, where));

	while (nbytes-- > 0) {
		*buf++ = *where++;
		bytes_read++;
	}
	f->pos += bytes_read;
	return bytes_read;
}

/*
 * shm_ioctl: currently, the only IOCTL's available are:
 * SHMSETBLK:  set the address of the shared memory file. This
 *             call may only be made once per region, and then only
 *	       if the region is open for writing.
 * SHMGETBLK:  get the address of the shared memory region. This
 *             call fails (returns 0) if SHMSETBLK has not been
 *             called yet for this shared memory file.
 */

static long _cdecl 
shm_ioctl(FILEPTR *f, int mode, void *buf)
{
	SHMFILE *s;
	MEMREGION *m;
	int i;
	long r;

	s = (SHMFILE *)f->devinfo;
	switch(mode) {
	case SHMSETBLK:
		if (s->reg) {
			DEBUG(("Fcntl: SHMSETBLK already performed for %s",
				s->filename));
			return ERANGE;
		}
		if ((f->flags & O_RWMODE) == O_RDONLY) {
			DEBUG(("Fcntl: SHMSETBLK: %s was opened read-only",
				s->filename));
			return EACCDN;
		}
	/* find the memory region to be attached */
		m = 0;
		for (i = curproc->num_reg - 1; i >= 0; i--) {
			if (curproc->addr[i] == (virtaddr)buf) {
				m = curproc->mem[i];
				break;
			}
		}
		if (!m || !buf) {
			DEBUG(("Fcntl: SHMSETBLK: bad address %lx", buf));
			return EIMBA;
		}
		if (m->shadow) {
			DEBUG(("Fcntl: SHMSETBLK: cannot share forked region"));
			return EACCDN;
		}
		m->mflags |= M_SHARED;
		m->links++;
		s->reg = m;
		return E_OK;
	case SHMGETBLK:
		if ((m = s->reg) == 0) {
			DEBUG(("Fcntl: no address for SHMGETBLK"));
			return E_OK;
		}
	/* check for memory limits */
		if (curproc->maxmem) {
			if (m->len > curproc->maxmem - memused(curproc)) {
				DEBUG(("Fcntl: SHMGETBLK would violate memory limits"));
				return E_OK;
			}
		}
		return (long)attach_region(curproc, m);
	case FIONREAD:
	case FIONWRITE:
		if (s->reg == 0) {
			r = 0;
		} else {
			r = s->reg->len - f->pos;
			if (r < 0) r = 0;
		}
		*((long *)buf) = r;
		return E_OK;
	default:
		DEBUG(("shmfs: bad Fcntl command"));
	}
	return EINVFN;
}

static long _cdecl 
shm_lseek(FILEPTR *f, long int where, int whence)
{
	long newpos, maxpos;
	SHMFILE *s;

	s = (SHMFILE *)f->devinfo;

	if (s->reg)
		maxpos = s->reg->len;
	else
		maxpos = 0;

	switch(whence) {
	case SEEK_SET:
		newpos = where;
		break;
	case SEEK_CUR:
		newpos = f->pos + where;
		break;
	case SEEK_END:
		newpos = maxpos + where;
		break;
	default:
		return EINVFN;
	}

	if (newpos < 0 || newpos > maxpos)
		return ERANGE;

	f->pos = newpos;
	return newpos;
}

static long _cdecl 
shm_datime(FILEPTR *f, ushort *timeptr, int rwflag)
{
	SHMFILE *s;

	s = (SHMFILE *)f->devinfo;
	if (rwflag) {
		s->time = *timeptr++;
		s->date = *timeptr;
	} else {
		*timeptr++ = s->time;
		*timeptr = s->date;
	}
	return E_OK;
}

static long _cdecl 
shm_close(FILEPTR *f, int pid)
{
	SHMFILE *s;

	UNUSED(pid);
	if (f->links <= 0) {
		s = (SHMFILE *)f->devinfo;
		s->inuse--;
	}
	return E_OK;
}
